(*
Author:  Akihisa Yamada (2017)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2017)
License: LGPL (see file COPYING.LESSER)
*)
theory Max_Polynomial_Impl
  imports
    Max_Polynomial
    Ordered_Algebra_Impl
    IA_Checker
    Term_Order_Impl
begin

locale max_poly_encoding_impl =
  verbose.encoder: encoding_impl
  where default_fun = "\<lambda>x. max_poly.MaxF" +
  verbose: pre_encoded_ord_algebra
  where less_eq = "(\<le>)" and less = "(<)" and I = max_poly.I and E = verbose.encoder.E
begin

text \<open>
  We simplify the encoding terms. It is shown to preserve semantics, but the fact doesn't matter
  for soundness.
\<close>
definition "simplified_alist = map (map_prod id max_poly.simplify) alist"

sublocale encoder: encoding_impl
  where alist = "simplified_alist"
    and default_fun = "\<lambda>_. max_poly.MaxF" .

sublocale pre_encoded_ord_algebra
  where less_eq = "(\<le>)"
    and less = "(<)"
    and I = max_poly.I
    and E = encoder.E .

definition "check_less_term h s t = IA.check_formula h (encode s \<sqsubset>\<^sub>I encode t)"

definition "check_less_eq_term h s t = IA.check_formula h (encode s \<sqsubseteq>\<^sub>I encode t)"

definition "check_ws_fun_arg h f n i = do {
  check (i < n) (showsl (STR ''bad argument number''));
  IA.check_formula h (Var i \<sqsubseteq>\<^sub>I encoder.E f n)
}"

definition [simp]: "create_max_poly_redtriple = \<lparr>
  redtriple.valid = encoder.check_encoding,
  s = (\<lambda>(s, t). check_less_term default t s),
  ns = (\<lambda>(s, t). check_less_eq_term default t s),
  nst = (\<lambda>(s, t). check_less_eq_term default t s),
  af = (\<lambda>(f, n). {0..<n} - set (constant_positions f n)),
  mono_af = empty_af,
  mono = (\<lambda> s_ns_nst. error (showsl_lit (STR ''monotonicity of max-poly is not yet supported''))),
  desc = showsl_lit (STR ''max-polynomial interpretations:'') \<circ>
    showsl_nl \<circ> encoder.showsl_encoding,
  not_ws_ns = Some (map fst simplified_alist),
  cpx = no_complexity_check
\<rparr>"

lemma af_compatible:
  "af_compatible (\<lambda>(f, n). {0..<n} - set (constant_positions f n)) (rel_of (\<sqsupseteq>\<^sub>t))"
 (is "af_compatible ?pi _")
proof -
  {
    fix f and s t :: "('a, 'b) term" and bef aft :: "('a, 'b) term list" 
    let ?ss = "bef @ s # aft" 
    let ?n = "Suc (length bef + length aft)" 
    assume "length bef \<notin> ?pi (f, ?n)" 
    then have "length bef \<in> set (constant_positions f ?n)" by auto
    from constant_atD [OF constant_positions [OF this], of "map (\<lambda>s. \<lbrakk>s\<rbrakk>\<alpha>) ?ss" "\<lbrakk>t\<rbrakk>\<alpha>" for \<alpha>, simplified]
    have "Fun f (bef @ s # aft) \<sqsupseteq>\<^sub>t Fun f (bef @ t # aft)"
      by (auto simp: nth_append list_update_append)
  }
  then show ?thesis unfolding af_compatible_def by blast
qed

context
  assumes ok: "isOK encoder.check_encoding"
begin

interpretation encoder: encoding where E = encoder.E by (fact encoder.check_encoding[OF ok])
interpretation max_poly_encoding where E = encoder.E ..

lemma check_less_term:
  assumes "isOK (check_less_term h s t)" shows "s \<sqsubset>\<^sub>t t"
  by (intro less_term_via_IA IA.check_formula, insert assms, unfold check_less_term_def, auto)

lemma check_less_eq_term:
  assumes "isOK (check_less_eq_term h s t)" shows "s \<sqsubseteq>\<^sub>t t"
  by (intro less_eq_term_via_IA IA.check_formula, insert assms, unfold check_less_eq_term_def, auto)

lemmas redpair = redpair

lemma check_ws_fun_arg:
  assumes "isOK (check_ws_fun_arg h f n i)" shows "ws_fun_arg (rel_of (\<sqsupseteq>\<^sub>t)) (f, n) i"
  using assms by (intro ws_fun_arg_via_IA IA.check_formula, auto simp: check_ws_fun_arg_def)

end

lemma default_E: "(f, n) \<notin> fst ` set simplified_alist \<Longrightarrow>
  encoder.E f n = Fun max_poly.MaxF (map Var [0..<n])"
  by (auto simp: encoder.E_def split: option.split dest!: map_of_SomeD)

end

lemma le_max_list: "(x::nat) \<in> set xs \<Longrightarrow> x \<le> max_list xs"
  using split_list by fastforce

lemma le_max_list1: "(x::nat) \<in> set xs \<Longrightarrow> x \<le> max_list1 xs"
  by (subst max_list_as_max_list1 [symmetric]) (auto simp: le_max_list)

lemma max_poly_redtriple: "generic_redtriple_impl max_poly_encoding_impl.create_max_poly_redtriple"
proof (unfold_locales, goal_cases)
  case (1 rp s_list ns_list nst_list)
  interpret max_poly_encoding_impl rp.
  from 1(1)[simplified] interpret encoder: encoding where E = encoder.E by (rule encoder.check_encoding)
  interpret max_poly_encoding where E = encoder.E..
  interpret redpair_order "rel_of (\<sqsupset>\<^sub>t)" "rel_of (\<sqsupseteq>\<^sub>t)"..
  interpret redtriple_order "rel_of (\<sqsupset>\<^sub>t)" "rel_of (\<sqsupseteq>\<^sub>t)" "rel_of (\<sqsupseteq>\<^sub>t)"
    apply (unfold_locales; (intro redpair.subst_NS term.trans_NS term.compat_NS_S term.refl_NS)?)
    using term.less_imp_le by auto
  note 1 = 1 [simplified]
  note 2 = check_less_eq_term [OF 1(1), of "default"] check_less_term [OF 1(1), of "default"]
  have 3: "set s_list \<subseteq> rel_of (\<sqsupset>\<^sub>t)"
    using 1(2) [rule_format] and 2(2)
    by auto
  have 4: "set ns_list \<subseteq> rel_of (\<sqsupseteq>\<^sub>t)"
    using 1(3)[rule_format] and 2(1)
    by auto
  have 5: "set nst_list \<subseteq> rel_of (\<sqsupseteq>\<^sub>t)"
    using 1(4)[rule_format] and 2(1)
    by auto
  have [simp]: "map fst rp = map fst simplified_alist" by (auto simp: simplified_alist_def)
  have [simp]: "map (snd \<circ> fst) rp = map (snd \<circ> fst) simplified_alist"
    unfolding map_map [symmetric] by simp
  define n where "n \<equiv> max_list1 (map (snd \<circ> fst) rp) + 1"
  then have m: "\<And>c m. m \<ge> n \<Longrightarrow> (c, Suc (Suc m)) \<notin> set (map fst rp)"
    apply (auto simp:)
    apply (subgoal_tac "Suc (Suc m) \<le> max_list1 (map (snd \<circ> fst) simplified_alist)")
     apply simp
    apply (rule le_max_list1, force)
    done
  have 6: "\<forall>m\<ge>n. \<forall>c. ce_trs (c, m) \<subseteq> rel_of (\<sqsupseteq>\<^sub>t)"
  proof -
    {
      fix m c and l r :: "('a, _) term"
      assume nm: "n \<le> m" and "(l, r) \<in> ce_trs (c,m)"
      then obtain s t where l: "l = Fun c ([s,t] @ replicate m (Var undefined))" (is "_ = Fun _ ?ls") and r: "r = s \<or> r = t" 
        unfolding ce_trs.simps by auto
      have len: "length ?ls = Suc (Suc m)" by simp
      have id: "[0..<Suc (Suc m)] = [0,1] @ map (Suc o Suc) [0 ..< m]"
        unfolding map_map [symmetric] and map_Suc_upt
        by (auto simp: upt_conv_Cons)
      from m [OF nm] have "\<forall> \<alpha>. \<exists> k.\<lbrakk>l\<rbrakk>\<alpha> = max (\<lbrakk>s\<rbrakk> \<alpha>) (max (\<lbrakk>t\<rbrakk>\<alpha>) k)"
        unfolding default_E [OF m [OF nm, simplified]]
          and id and l and eval.simps and len and length_map
        by (auto)
      from choice [OF this] obtain k where "\<lbrakk>l\<rbrakk>\<alpha> = max (\<lbrakk>s\<rbrakk>\<alpha>) (max (\<lbrakk>t\<rbrakk>\<alpha>) (k \<alpha>))" for \<alpha> by blast
      with r have "l \<sqsupseteq>\<^sub>t r" by (auto intro: max.cobounded1)
      then have "(l,r) \<in> rel_of (\<sqsupseteq>\<^sub>t)" by auto
    }
    then show ?thesis by auto
  qed
  show ?case
    apply (rule exI[of _ "rel_of (\<sqsupset>\<^sub>t)"], intro exI[of _ "rel_of (\<sqsupseteq>\<^sub>t)"] conjI, unfold_locales)
    by (auto intro!: 3 4 5 6 empty_af af_compatible 
        simp:ws_fun_arg_def no_complexity_check_def o_def nth_map 
        dest!:default_E 
        intro!: less_eq_termI le_max_list)
qed

declare max_poly_encoding_impl.check_less_term_def [code]
declare max_poly_encoding_impl.check_less_eq_term_def [code]
declare max_poly_encoding_impl.simplified_alist_def [code]
declare max_poly_encoding_impl.create_max_poly_redtriple_def [code]
declare pre_encoded_algebra.constant_positions_def [code]

end
